/*
        cobr2tap tool by Symoon (May 2022)
	based on wav2tap 1.2 and wav2taps tools, by F.Frances and Symoon 

        corb2tap is designed to convert WAVs of Cobra Soft protected tapes.
        Cobra Soft sometimes used special routines mixing fast and slow signal, and programs without headers.

        WAV file must be of good quality, in 44kHz mono 8-bit format, then processed with Wavclean 1.2 or further.
        It's much better to remove all silences in the initial WAV file, between the different parts
        and at the end, to avoid trash bytes from trying to decode silence or parasites between each part.

        With thanks to Dom for his detailed study of the Cobra Soft routines.
*/


#include <stdio.h>

struct {
	char sig[4];
	int riff_size;
	char datasig[4];
	char fmtsig[4];
	int fmtsize;
	short tag;
	short channels;
	int freq;
	int bytes_per_sec;
	short byte_per_sample;
	short bits_per_sample;
	char samplesig[4];
	int length;
} sample_riff;

int lengthbit;
int err_bit=0; //indique si le bit lu prcdemment tait en erreur (donc probablement synchro  refaire) (1=oui)
int sync_ok=0;
int read_sync=1;
int offset;
long total_slow_bytes=0;
FILE *in, *out;


init(int argc, char *argv[])
{
	if (argc<3) return 1;
	if (argv[1][0]=='-') {
		if (feed_parameters(argv[1])) return 1;
	}

	in=fopen(argv[argc-2],"rb");
	if (in==NULL) {
		printf("Cannot open %s file\n\n",argv[argc-2]);
		return 1;
	}

	fread(&sample_riff,sizeof(sample_riff),1,in);
	if (sample_riff.channels!=1 || sample_riff.freq!=4800 || sample_riff.byte_per_sample!=1) {
		printf("Invalid WAV format: should be 4800 Hz, 8-bit, mono\n");
		return 1;
	}

	out=fopen(argv[argc-1],"wb");
	if (out==NULL) {
		printf("Cannot create %s file\n\n",argv[argc-1]);
		return 1;
	}
	return 0;
}

int feed_parameters(char *param)
{
	switch (param[1]) {
		case 'r':
		case 'R':
			read_sync = 0;
			break;
		default:
			printf("Bad parameter: %s\n", param);
			return 1;
	}
	return 0;
}


main(int argc,char **argv)
{	
	unsigned start,end,byte;
	if (init(argc,argv)) {
		printf("\nCobr2tap 1.0 (May 2022)\n\nThis program decodes Cobra Soft protected programs, replacing\n");
                printf("Wav2tap if it failed (Cobra Soft protection mixes FAST and SLOW speed)\n");
                printf("Beware that all Cobra Soft tapes are not protected, so try Wav2tap first!\n\n");
                printf("Usage:\n1- convert your mono 8-bit 44kHz WAV file with Wavclean 1.2 or further\n");
                printf("2- convert the resulting new WAV file with cobr2tap file.wav file.tap\n\n",argv[0]);

		printf("Options: -r  tries to read without synchro. Useful to look what are these\n");
		printf("             programs that were partially erased on tape, or that have\n");
		printf("             non-standard synchro, but will probably lose a few bytes at\n");
		printf("             the beginning.\n");
		printf("             (default will require a synchro first)\n");
		exit(1);}

	for (;;) {
		if (read_sync) {
		synchronize();
		printf("Synchro found, ");
		}

		printf("decoding bytes...\n");
		putc(0x16,out); putc(0x16,out); putc(0x16,out);
		sync_ok=1;
		offset+=3;
//		byte=0x24;              // 0x24 removed as it's not always the starting byte with Cobra Soft
		while (sync_ok) {
			byte=getbyte();
			putc(byte,out);
			offset++;
		}
	}
}

getbit()
{
	int val;

	lengthbit=1;
	val=getc(in);
	if (val==EOF) sortie();
	while (val>=0x80) {
		lengthbit++;
		val=getc(in);
		if (val==EOF) sortie();
	}
	lengthbit++;
	val=getc(in);
	if (val==EOF) sortie();
	while (val<=0x80) {
		lengthbit++;
		val=getc(in);
		if (val==EOF) sortie();
	}
	if (lengthbit>10) return -1;
	else if (lengthbit>2) return 0;
	else return 1;
}

getbitslow()
{
	//Stratgie: lire 3 octets. Si le 1 est gal au 3, c'est un "1", sinon c'est un "0"
        //  C'est le tableau VALREAD qui stocke ces 3 octets et etablit cette valeur.
	//  16 octets composant un bit, on boucle 14 fois pour vrifier la valeur du bit
	//  (une lecture des 3 premiers octets + 13 lectures des octets restants)
        //  Le tableau TABREAD reprend l'ensemble des valeurs lues, qui sont donc en
        //  principe toutes les memes.
	//
        //
        // ATTENTION: le signal issu d'anciennes versions de WAVCLEAN donne, pour 0,
        //            une forme __-__-__-__- au lieu de __--__--__--__--
	//            Corrig a partir de Wavclean 1.2.
        // RAPPEL: nombre aleatoire de periodes a la fin de l'en-tete apres le nom !
        //         il faut se positionner aprs le 00 qui termine le nom,  savoir:
        //         aprs le #24, 9 octets aprs, au premier octet valant 00
        //         boucler tant qu'on dtecte des bits 1. Attention le nombre de
        //         priodes peut tre incohrent par rapport aux bits => dtecter
        //         et tester ds qu'on tombe sur un bit 0, et se re-synchroniser...

	int valread[3],valbit=0,i,tabread[14]={0,0,0,0,0,0,0,0,0,0,0,0,0,0};
	char waveform[]="________________"; // les 16 demi-priodes du signal



	// Premire lecture - galits parfaites avec (0x80) des  Wavclean qui centre le signal
	valread[0]=getc(in);
	if (valread[0]==EOF) sortie();
	if (valread[0]>=0x80) waveform[0]='-';
	valread[1]=getc(in);
	if (valread[1]==EOF) sortie();
	if (valread[1]>=0x80) waveform[1]='-';
	valread[2]=getc(in);
	if (valread[2]==EOF) sortie();
	if (valread[2]>=0x80) waveform[2]='-';
	if (valread[0]==valread[2])
	{	valbit++;
		tabread[0]=1;
	}

	// Lectures suivantes
	for (i=4;i<=16;i++)
	{	valread[0]=valread[1];
		valread[1]=valread[2];
		valread[2]=getc(in);
		if (valread[2]==EOF) sortie();
		if (valread[2]>=0x80) waveform[i-1]='-';
		if (valread[0]==valread[2])
		{	valbit++;
			tabread[i-3]=1;
		}
	}

        //printf("bit at offset $%x, read %d  %s \n",offset,valbit,waveform);

	// --- Analyse du bit lu ---
	// Si on a lu 14 fois "0", tout va bien: bit = 0
	// On a lu 14 fois "1", tout va bien: bit = 1
	// Sinon on a lu des valeurs diffrentes pour le mme bit, donc il faut analyser:
	// - si on a lu "1" au moins 7 fois, on retourne 1
	// - sinon on retourne 0

	if (valbit==0) return 0;
	else if (valbit==14) return 1; 
	else {
		err_bit=1; //consider being desynchronized because of error
		long position_fich=ftell(in); // position in WAV source file
		if (valbit>=7) 
		//si on n'est pas en mode display, on n'affiche les erreurs que si on est aprs la synchro
		{ if (offset>0) printf("bit error at offset $%x, read %s, returned 1. WAV pos $%x\n",offset,waveform, position_fich);
		  return 1;
		}
		else
		{ if (offset>0) printf("bit error at offset $%x, read %s, returned 0. WAV pos $%x\n",offset,waveform, position_fich);
                  return 0;
		}
	      }
}


getbyte()
{
	int decaleur=0,byte=0,i,bit,sum=0, first_zero_length=0, additional_length=0;
        int slowbyte=0; // 1=slow, other=fast
        int garbageread[2];
	getbit();
	while((bit=getbit())==1); // fast reading stop bits works with both fast and slow as stop bits have the same format.
	if (bit==-1) { sync_ok=0; return 0; }

        long position_fich=ftell(in); // position of the byte data in WAV source file (or in the start bit for slow byte)

        // right here, we have read a zero bit corresponding to the start bit.
        // Its lentgh is 3 for a fast byte, or 4 followed by 3 more identical lengths for a slow byte.
        // So we'll read 3 more bits, and check to total length to confirm it's a slow bit, then go back to carry on reading.
        // For a slow bit, the total length should be 4*4 = 16
        // For a fast bit, after the start bit, we would read 3 more bits of lentgh 2 or 3, so between 3+3*2=9 and 3+3*3=12
        // => So as there could be errors or half-bits, let's consider we have a slow bit if length >=14 (insteadd of 16)

        first_zero_length=lengthbit;
	for(i=0;i<3;i++)
        { getbit();
          additional_length=additional_length+lengthbit;
        }
        if((first_zero_length+additional_length)>=14) slowbyte=1;

        if(slowbyte) // ******* slow byte *******
        { // search the real start of the start byte (thus skipping half-period from Oric ROM bug)
          // => go back to zero
          fseek(in,0-(first_zero_length+additional_length+3),1);  //+3 can't hurt, some bad readings didn't bring back enough.

          total_slow_bytes++; // check to warn the user if no slow bytes detected: better use WAV2TAP.

          //printf("S $%x\n",offset);
          // read until two values are equal
          garbageread[0]=getc(in);
          garbageread[1]=getc(in);

          while (garbageread[0]!=garbageread[1])
          { garbageread[0]=garbageread[1];
            garbageread[1]=getc(in);
          }
          // Once equal, we're at the start of the zero start bit so go back two positions earlier
          fseek(in,0-2,1);

          // Now skip the start bit
          getbitslow();

          // Now we can read the slow byte
	  for(i=0;i<8;i++) decaleur=(decaleur<<1)|getbitslow();
	  for(i=0;i<8;i++) {
	    bit=decaleur&1;
	    decaleur=decaleur>>1;
	    byte=(byte<<1)|bit;
	    sum+=bit;
          }
	  if (((sum&1)==getbitslow()) && sync_ok) printf("parity error at offset $%x (slow byte), WAV pos $%x\n",offset,position_fich);
	  return byte;
          // bit errors after slow byte is neglected (no resynchronization) as the start synchro is done
          // with fast bits, and a slow byte ends with 24 fast bits, so that should be enough to recover!
        }

        else // ******* fast byte *******
        { // search the real start of the start byte (thus skipping half-period from Oric ROM bug)
          // => go back after start bit
          fseek(in,0-(additional_length),1);

          // Now we can read the fast byte
	  for(i=0;i<8;i++) decaleur=(decaleur<<1)|getbit();
	  for(i=0;i<8;i++) {
		bit=decaleur&1;
		decaleur=decaleur>>1;
		byte=(byte<<1)|bit;
		sum+=bit;
	  }
	  if (((sum&1)==getbit()) && sync_ok) printf("parity error at offset $%x (fast byte), WAV pos $%x\n",offset,position_fich);
	  return byte;
        }
}

synchronize()
{
	int decaleur=0;
	printf("Searching synchro...\n");
	while(1) {
		while((decaleur&0xff) != 0x68)
                        decaleur=(decaleur<<1)|getbit();
		if (getbyte()!=0x16) continue;
		if (getbyte()!=0x16) continue;
		if (getbyte()!=0x16) continue;
		break;
	}
//	while (getbyte()!=0x24);  // 0x24 removed as it's not always the starting byte with Cobra Soft
}

sortie()
{
        printf("%ld slow bytes were found.\n",total_slow_bytes);
        if (total_slow_bytes<5) printf("That REALLY wasn't much, probably not Cobra Soft protected, try WAV2TAP!\n");
        exit(0);
}
